   
   
   
   MODULE Process_ISH_Subs
      USE GlerlUtil
      USE MetDataTypesAndUnits
      USE DailyMetStationDataset
      USE ReadWriteMetCsv

      TYPE ObsTimeValue
         INTEGER :: ObsDate             !  YYYYMMDD as a single integer value
         INTEGER :: ObsTime             !  HHMM as a single integer value
         REAL    :: DVal
      END TYPE ObsTimeValue

      TYPE PrecipObs
         INTEGER :: ObsDate             !  YYYYMMDD as a single integer value
         INTEGER :: ObsTime             !  HHMM as a single integer value
         INTEGER :: Hours               !  Hours of accumulation represented by the value
         REAL    :: DVal                !  millimeters
      END TYPE PrecipObs

      TYPE DataFromOneLine
         CHARACTER (LEN=6) :: USAF
         CHARACTER (LEN=5) :: WBAN
         INTEGER :: ObsDate             !  YYYYMMDD as a single integer value
         INTEGER :: ObsTime             !  HHMM as a single integer value
         INTEGER :: DateSeq             !  Date as a sequence number
         INTEGER :: Time                !  HHMM as a single integer
         REAL    :: Lat                 !  latitude in decimal degrees north of equator
         REAL    :: Long                !  longitude in decimal degrees east of prime meridian
         REAL    :: AirTemp             !  degrees C
         REAL    :: DewPoint            !  degrees C
         REAL    :: WindSpd             !  meters/sec
         REAL    :: Cloud               !  Percent of sky covered by cloud (distilled from detailed info)
         TYPE (PrecipObs), DIMENSION(4) :: Precip
      END TYPE DataFromOneLine


      
   CONTAINS
   
   
!--------------------------------------------------------------------
      SUBROUTINE ParseDataLine(Line, DFOL, ErrStat)
      IMPLICIT NONE
      CHARACTER(LEN=*),       INTENT(IN)    :: Line
      TYPE (DataFromOneLine), INTENT(INOUT) :: DFOL
      INTEGER,                INTENT(OUT)   :: ErrStat
      
      INTEGER :: I, J, IOS, VarChars, YYYY, MM, DD, HH, NN, Seq
      INTEGER :: Hrs, SkyCode1, SkyCode2, SkyQC, SkyPct(6)
      LOGICAL :: OK
      REAL    :: Prc
      CHARACTER(LEN=1)    :: CC, QC, AirQC, DewQC, WindQC
      CHARACTER(LEN=3)    :: I3
      CHARACTER(LEN=8)    :: S8
      CHARACTER(LEN=12)   :: S12
      CHARACTER(LEN=9999) :: VLine
      
      !
      !  Initialize the DFOL fields to missing data
      !
      CALL ClearDFOL(DFOL)
      
      !
      !  Read the Mandatory data section
      !
      READ(Line, 1000, IOSTAT=IOS) VarChars, DFOL%USAF, DFOL%WBAN,        &
                 YYYY, MM, DD, HH, NN,                                    &
                 DFOL%Lat, DFOL%Long, DFOL%WindSpd, WindQC,               &
                 DFOL%AirTemp, AirQC, DFOL%DewPoint, DewQC
      IF (IOS .NE. 0) THEN
         CALL ClearDFOL(DFOL)
         ErrStat = 1
         RETURN
      END IF

      !
      !  Adjust to local time from UTC (simple 5-hour shift for now)
      !  
      CALL TranslateUTC_Local(YYYY, MM, DD, HH)
      
      !
      !  Build the big date and time integer values
      !
      
      DFOL%ObsDate = (YYYY * 10000) + (MM * 100) + DD 
      DFOL%ObsTime = (HH * 100) + NN
                     
      !
      !  Fill the date/time fields in DFOL
      !
      CALL DateSequence(DD, MM, YYYY, DFOL%DateSeq)
      IF (ErrorLevel .NE. 0) THEN
         CALL ClearDFOL(DFOL)
         ErrStat = 2
         RETURN
      END IF
      
      !
      !  If the data value is the "Missing" value, replace it with our
      !  defined MissingData value
      !
      IF (DFOL%WindSpd  .GT. 999.7) DFOL%WindSpd = MissingData_Real
      IF (DFOL%AirTemp  .GT. 999.7) DFOL%AirTemp = MissingData_Real
      IF (DFOL%DewPoint .GT. 999.7) DFOL%DewPoint = MissingData_Real
      
      !
      !  Inspect the quality code for air temp and use it to filter 
      !  the airtemp values.
      !
      OK = .TRUE.
      IF (AirQC .EQ. '2') OK = .FALSE.
      IF (AirQC .EQ. '3') OK = .FALSE.
      IF (AirQC .EQ. '6') OK = .FALSE.
      IF (AirQC .EQ. '7') OK = .FALSE.
      IF (.NOT. OK) DFOL%AirTemp = MissingData_Real
      
               
      !
      !  Inspect the quality code for dewpt temp and use it to filter 
      !  the dewpt values.
      !
      OK = .TRUE.
      IF (DewQC .EQ. '2') OK = .FALSE.
      IF (DewQC .EQ. '3') OK = .FALSE.
      IF (DewQC .EQ. '6') OK = .FALSE.
      IF (DewQC .EQ. '7') OK = .FALSE.
      IF (.NOT. OK) DFOL%Dewpoint = MissingData_Real
               
      !
      !  Inspect the quality code for windspeed and use it to filter 
      !  the windspeed value.
      !
      OK = .TRUE.
      IF (WindQC .EQ. '2') OK = .FALSE.
      IF (WindQC .EQ. '3') OK = .FALSE.
      IF (WindQC .EQ. '6') OK = .FALSE.
      IF (WindQC .EQ. '7') OK = .FALSE.
      IF (.NOT. OK) DFOL%WindSpd = MissingData_Real
               
      !
      !  Extract the additional data into a separate string variable
      !      
      J = 105 + VarChars
      VLine = TRIM(Line(106:J))
                     
      !
      !  Search the additional data section for precip entries
      !
      DO I = 1, 4
         WRITE(I3, 1500) 'AA', I      ! 3-char identifier for the field
         J = INDEX(VLine, I3)
         IF (J .GT. 0) THEN
            S8 = VLine(J+3:J+10)
            READ(S8, 1020, IOSTAT=IOS) Hrs, Prc, CC, QC 
            IF (Prc .GT. 999.7) Prc = MissingData_Real      ! missing data value
            IF (IOS .EQ. 0) THEN
               DFOL%Precip(I)%Hours = Hrs
               DFOL%Precip(I)%DVal  = Prc
               OK = .TRUE.
               
               !
               !  The "Condition Code" (or trace indicator) is a bit confusing.
               !  In particular, the Begin/End period entries seem to make no
               !  sense, because they don't "book end". I asked a guy at NCDC 
               !  about these and he was unable to provide any insight.
               !  Therefore, I am going to ignore the Begin/End entries.
               !  If the value is __ I do __ with the precip data.
               !   '1' (Measurement impossible or inaccurate)   : throw it away
               !   '2' (Trace)                                  : keep it
               !   '3' - '8' (those Begin/End options           : ignore this field
               !   'E' (Estimated, e.g. from nearby station)    : keep it (reluctantly)
               !   'I' (Incomplete, excludes missing reports)   : throw it away
               !   'J' (Incomplete, excludes erroneous reports) : throw it away
               !   '9' (Missing)                                : ignore this field
               !
               IF (CC .EQ. '1') OK = .FALSE.
               IF (CC .EQ. 'I') OK = .FALSE.
               IF (CC .EQ. 'i') OK = .FALSE.
               IF (CC .EQ. 'J') OK = .FALSE.
               IF (CC .EQ. 'j') OK = .FALSE.
               
               !
               !  The Quality Code is used/interpreted as follows:
               !   '0'   :  keep
               !   '1'   :  keep
               !   '2'   :  throw away
               !   '3'   :  throw away
               !   '4'   :  keep
               !   '5'   :  keep
               !   '6'   :  throw away
               !   '7'   :  throw away
               !   '9'   :  keep
               !   'A'   :  keep
               !   'I'   :  keep
               !   'M'   :  keep
               !   'P'   :  keep
               !   'R'   :  keep
               !   'U'   :  keep
               !
               IF (QC .EQ. '2') OK = .FALSE.
               IF (QC .EQ. '3') OK = .FALSE.
               IF (QC .EQ. '6') OK = .FALSE.
               IF (QC .EQ. '7') OK = .FALSE.
               
               IF (.NOT. OK) THEN
                  DFOL%Precip(I)%Hours = -99
                  DFOL%Precip(I)%DVal  = MissingData_Real
               END IF
            END IF
         END IF
      END DO
                     
      !
      !  Search the additional data section for 'sky-cover-summation-state' entries.
      !  I will ignore the height information since all I care about is sky cover.
      !  Look at SkyCode1 first. Then SkyCode2. That way we preferentially use SkyCode2.
      !
      !  For SkyCode1 I will only use values 0-4.
      !     5 and 6 are descriptive values that I cannot interpret to a %
      !     9 = missing
      !
      !  For SkyCode2 I will only use the values from 0-8. 
      !     9     = 'unable to estimate'
      !     10-19 = descriptive stuff that I cannot interpret into a %
      !     99    = missing
      !
      DO I = 1, 6
         SkyPct(I) = -99
         WRITE(I3, 1500) 'GD', I      ! 3-char identifier for the field
         J = INDEX(VLine, I3)
         IF (J .GT. 0) THEN
            S12 = VLine(J+3:J+14)
            READ(S12, 1030, IOSTAT=IOS) SkyCode1, SkyCode2, SkyQC
            IF (IOS .EQ. 0) THEN
               IF (SkyCode1 .EQ. 0) SkyPct(I) = 0
               IF (SkyCode1 .EQ. 1) SkyPct(I) = 13  ! halfway between 1 and 25
               IF (SkyCode1 .EQ. 1) SkyPct(I) = 43  ! halfway between 37 and 50
               IF (SkyCode1 .EQ. 1) SkyPct(I) = 69  ! halfway between 50 and 88
               IF (SkyCode1 .EQ. 1) SkyPct(I) = 100

               IF (SkyCode2 .EQ. 0) SkyPct(I) = 0
               IF (SkyCode2 .EQ. 1) SkyPct(I) = 5
               IF (SkyCode2 .EQ. 2) SkyPct(I) = 25
               IF (SkyCode2 .EQ. 3) SkyPct(I) = 40
               IF (SkyCode2 .EQ. 4) SkyPct(I) = 50
               IF (SkyCode2 .EQ. 5) SkyPct(I) = 60
               IF (SkyCode2 .EQ. 6) SkyPct(I) = 75
               IF (SkyCode2 .EQ. 7) SkyPct(I) = 95
               IF (SkyCode2 .EQ. 8) SkyPct(I) = 100

               !
               !  The Quality Code is used/interpreted as follows:
               !   '0'   :  keep
               !   '1'   :  keep
               !   '2'   :  throw away
               !   '3'   :  throw away
               !   '4'   :  keep
               !   '5'   :  keep
               !   '6'   :  throw away
               !   '7'   :  throw away
               !   '9'   :  keep
               !
               IF (QC .EQ. '2') OK = .FALSE.
               IF (QC .EQ. '3') OK = .FALSE.
               IF (QC .EQ. '6') OK = .FALSE.
               IF (QC .EQ. '7') OK = .FALSE.
               
               IF (.NOT. OK) THEN
                  SkyPct(I) = -99
               END IF
            END IF
         END IF
      END DO
      
      !
      !  Look at all of the cloud layers. Use the max value as the overall coverage.
      !  This is the approach that Greg Lang said they use.
      !
      DFOL%Cloud = MissingData_Real        ! very large negative number
      DO I = 1, 6
         IF (SkyPct(I) .GE. 0) THEN
            IF (SkyPct(I) .GT. DFOL%Cloud) DFOL%Cloud = SkyPct(I)
         END IF
      END DO
      IF (DFOL%Cloud .LT. 0) DFOL%Cloud = MissingData_Real
     
      RETURN

     
   
 1000 FORMAT(I4, A6, A5, I4, 4I2, 1X, F6.3, F7.3, T66, F4.1, A1, T88, 2(F5.1,A1)) 
 1020 FORMAT(I2, F4.1, A1, A1)
 1030 FORMAT(I1, I2, I1)
 
 1500 FORMAT(A2, I1)
 1505 FORMAT(I4.4, 4I2.2)
   
      END SUBROUTINE ParseDataLine


!--------------------------------------------------------------------
      SUBROUTINE ClearDFOL(DFOL)
      IMPLICIT NONE
      TYPE (DataFromOneLine), INTENT(INOUT) :: DFOL
      INTEGER :: I
      DFOL%USAF     = '------'
      DFOL%WBAN     = '-----'
      DFOL%ObsDate  = MissingData_Int
      DFOL%ObsTime  = MissingData_Int
      DFOL%DateSeq  = MissingData_Int
      DFOL%Time     = MissingData_Int
      DFOL%Lat      = MissingData_Real
      DFOL%Long     = MissingData_Real
      DFOL%AirTemp  = MissingData_Real
      DFOL%Dewpoint = MissingData_Real
      DFOL%WindSpd  = MissingData_Real
      DFOL%Cloud    = MissingData_Real
      DO I=1,4
         DFOL%Precip(I)%ObsDate = MissingData_Int
         DFOL%Precip(I)%ObsTime = MissingData_Int
         DFOL%Precip(I)%Hours   = -99
         DFOL%Precip(I)%DVal    = MissingData_Real
      END DO
      
      END SUBROUTINE ClearDFOL

      
!--------------------------------------------------------------------
!  The time in the file is UTC. 
!  For simplicity of all subsequent operations, I want to adjust 
!  the date/time fields from UTC to local time. 
!
!  Since I am going to be processing data for the Great Lakes, and
!  most of the G.L. basin is Eastern time, I am gonna make a fairly
!  big simplifying assumption -- 
!  ALL stations are converted from UTC to Eastern Standard Time.
!  I'm not even gonna worry about daylight savings time for now.
!  I will just adjust everything by 5 hours.
!
      SUBROUTINE TranslateUTC_Local(YYYY, MM, DD, HH)
      IMPLICIT NONE
      INTEGER, INTENT(INOUT) :: YYYY, MM, DD, HH
      INTEGER :: Seq
      
      HH = HH - 5
      IF (HH .LT. 0) THEN
         HH = HH + 24
         DD = DD - 1
         IF (DD .LT. 1) THEN
            CALL DateSequence(DD+1, MM, YYYY, Seq)
            CALL SequenceDate(DD, MM, YYYY, Seq-1)
         END IF
      END IF
      RETURN
      END SUBROUTINE TranslateUTC_Local


!--------------------------------------------------------------------
      SUBROUTINE ComputeDailyMax(RawData, DlyData, SeqStart, SeqFinish)
      IMPLICIT NONE
      TYPE (ObsTimeValue), DIMENSION(:), INTENT(IN)  :: RawData
      REAL,                DIMENSION(:), INTENT(OUT) :: DlyData
      INTEGER, INTENT(IN)  :: SeqStart, SeqFinish
      
      INTEGER :: I, Y, M, D, Seq, DDate
      INTEGER :: DayNum, NumEntries
      REAL    :: DVal
      CHARACTER(LEN=8) :: ST
      
      DlyData(:) = MissingData_Real
      
      NumEntries = UBOUND(RawData, 1)
      DO I = 1, NumEntries
         DDate = RawData(I)%ObsDate
         DVal  = RawData(I)%DVal
         IF ((DDate .GT. 0) .AND. (DVal .GT. MissingData_Real_Test)) THEN
            WRITE(ST, 1000) RawData(I)%ObsDate
            READ(ST, 1001) Y, M, D
            CALL DateSequence(D, M, Y, Seq)
            IF ((Seq .GE. SeqStart) .AND. (Seq .LE. SeqFinish)) THEN
               DayNum = Seq - SeqStart + 1
               IF (DlyData(DayNum) .LE. MissingData_Real_Test) DlyData(DayNum) = DVal
               IF (DVal .GT. DlyData(DayNum)) DlyData(DayNum) = DVal
            END IF
         END IF
      END DO
      
  1000 FORMAT(I8.8)
  1001 FORMAT(I4, 2I2)
   
      END SUBROUTINE ComputeDailyMax
  
!--------------------------------------------------------------------
      SUBROUTINE ComputeDailyMin(RawData, DlyData, SeqStart, SeqFinish)
      IMPLICIT NONE
      TYPE (ObsTimeValue), DIMENSION(:), INTENT(IN)  :: RawData
      REAL,                DIMENSION(:), INTENT(OUT) :: DlyData
      INTEGER, INTENT(IN)  :: SeqStart, SeqFinish
      
      INTEGER :: I, Y, M, D, Seq, DDate
      INTEGER :: DayNum, NumEntries
      REAL    :: DVal
      CHARACTER(LEN=8) :: ST
      
      DlyData(:) = MissingData_Real
      
      NumEntries = UBOUND(RawData, 1)
      DO I = 1, NumEntries
         DDate = RawData(I)%ObsDate
         DVal  = RawData(I)%DVal
         IF ((DDate .GT. 0) .AND. (DVal .GT. MissingData_Real_Test)) THEN
            WRITE(ST, 1000) RawData(I)%ObsDate
            READ(ST, 1001) Y, M, D
            CALL DateSequence(D, M, Y, Seq)
            IF ((Seq .GE. SeqStart) .AND. (Seq .LE. SeqFinish)) THEN
               DayNum = Seq - SeqStart + 1
               IF (DlyData(DayNum) .LE. MissingData_Real_Test) DlyData(DayNum) = DVal
               IF (DVal .LT. DlyData(DayNum)) DlyData(DayNum) = DVal
            END IF
         END IF
      END DO
      
  1000 FORMAT(I8.8)
  1001 FORMAT(I4, 2I2)
   
      END SUBROUTINE ComputeDailyMin
  
!--------------------------------------------------------------------
!  Doing daily max and min are pretty easy. Daily average is more involved because
!  There could potentially be a lot of clustered observations that can throw off 
!  the average.  For example, if we are computing a daily average air temp, and
!  the raw data has 10 mid-day observations, but only 1 night-time obs, that will
!  likely bias the average daily temp by a significant amount. In order to mitigate
!  that problem, I will first dump the raw data into hourly bins. Then compute a
!  simple arithmetic mean for each hour. Then compute simple arithmetic mean from
!  the (1-24) hourly values.
!
      SUBROUTINE ComputeDailyAvg(RawData, DlyData, SeqStart, SeqFinish)
      IMPLICIT NONE
      TYPE (ObsTimeValue), DIMENSION(:), INTENT(IN)  :: RawData
      REAL,                DIMENSION(:), INTENT(OUT) :: DlyData
      INTEGER, INTENT(IN)  :: SeqStart, SeqFinish
      
      INTEGER :: I, Y, M, D, H, Seq, DDate, DCount, IOS
      INTEGER :: DayNum, NumEntries, NumDays
      INTEGER :: DlyCount
      REAL    :: DVal, DlyAccum, HlyMean
      CHARACTER(LEN=8) :: ST
      
      INTEGER, DIMENSION(:,:), ALLOCATABLE :: HlyCount     !  dim(1:NumDays, 0:23)
      REAL,    DIMENSION(:,:), ALLOCATABLE :: HlyAccum     !  dim(1:NumDays, 0:23)
      
      
      DlyData(:) = MissingData_Real
      
      !
      !  Allocate the RAM for accumulating raw data into hourly bins
      !
      NumDays = SeqFinish - SeqStart + 1
      ALLOCATE(HlyAccum(NumDays, 0:23), HlyCount(NumDays, 0:23), STAT=IOS)
      IF (IOS .NE. 0) THEN
         PRINT*, 'Error allocating RAM for hourly bins of raw data.'
         CALL EXIT(1)
      END IF
      HlyAccum(:,:) = 0.0
      HlyCount(:,:) = 0

      !      
      !  Accumulate the raw data into those hourly bins
      !
      NumEntries = UBOUND(RawData, 1)
      DO I = 1, NumEntries
         DDate = RawData(I)%ObsDate
         DVal  = RawData(I)%DVal
         IF ((DDate .GT. 0) .AND. (DVal .GT. MissingData_Real_Test)) THEN
            WRITE(ST, 1000) RawData(I)%ObsDate
            READ(ST, 1001) Y, M, D
            WRITE(ST, 1004) RawData(I)%ObsTime
            READ(ST, 1005) H
            CALL DateSequence(D, M, Y, Seq)
            IF ((Seq .GE. SeqStart) .AND. (Seq .LE. SeqFinish)) THEN
               DayNum = Seq - SeqStart + 1
               HlyAccum(DayNum, H) = HlyAccum(DayNum, H) + DVal
               HlyCount(DayNum, H) = HlyCount(DayNum, H) + 1
            END IF
         END IF
      END DO
      
      !
      !  For each day...
      !    1. Compute the average value for each of the 24 hourly bins
      !    2. Compute daily average from those 24 values
      !
      DO DayNum = 1, NumDays
         DlyAccum = 0.0
         DlyCount = 0
         DO H = 0, 23
            IF (HlyCount(DayNum,H) .GT. 0) THEN
               HlyMean = HlyAccum(DayNum,H) / HlyCount(DayNum,H)
               DlyAccum = DlyAccum + HlyMean
               DlyCount = DlyCount + 1
            END IF
         END DO
         IF (DlyCount .GT. 0) THEN
            DlyData(DayNum) = DlyAccum / DlyCount
         END IF
      END DO
      
      DEALLOCATE(HlyAccum, HlyCount, STAT=IOS)      
      RETURN      
      
      
  1000 FORMAT(I8.8)
  1001 FORMAT(I4, 2I2)
  1004 FORMAT(I4.4, 4X)
  1005 FORMAT(I2)
   
      END SUBROUTINE ComputeDailyAvg
  

!--------------------------------------------------------------------
!  Computing daily precip is a beast due to the fact that the observations 
!  themselves are accumulations over varying periods of time.  I was curious
!  how NCDC computes daily precip from these various hourly data, so I asked.
!  The person I talked to had no answer for me. When I described the following
!  proposed algorithm he said that it sounded reasonable to him.  On the basis
!  of that ringing endorsement, I am proceeding as follows (haha):
!
!  1. Build a 1-d array representing every hour of the entire data period.
!  2. Step through the raw data, and every time I encounter a 1-hour precip
!     accumulation, assign the data value to that hour in the big 1-d array.
!  3. Step through, looking for any 2-hour accumulations. When found...
!        a. Look at the current and previous hours and see if one of them has 
!           already been assigned a value. If so, subtract that amount of precip
!           from the 2-hour accumulation. The remaining amount (if positive) is 
!           then assumed to have fallen during the other hour, so I assign it 
!           there.  If the remainder is < 0 then I assume some sort of roundoff
!           issue and just assign as if it were zero. Perhaps an "iffy" thing
!           to do, but I don't have any better idea.
!        b. If both of those hours have been assigned, just ignore the 2-hour
!           accumulation. Also an "iffy" option, but I don't have a better idea.
!        c. If neither of the hours has a value, equally apportion the accumulated
!           precip to those 2 hourly bins.
!   4. Do the same thing as number 3, but now for the 3-hour accumulations.
!   5. Do it for 4-hour accumulations.
!   6+ Do it over and over for each successive N-hour accumulation up to 48 hours.
!
!--------------------------------------------------------------------
      SUBROUTINE ComputeDailyPrecip(RawData, DlyData, SeqStart, SeqFinish)
      IMPLICIT NONE
      TYPE (PrecipObs), DIMENSION(:,:), INTENT(IN)  :: RawData
      REAL,             DIMENSION(:),   INTENT(OUT) :: DlyData
      INTEGER, INTENT(IN)  :: SeqStart, SeqFinish
      
      INTEGER :: I, J, IOS, AH, H1, H2, Y, M, D, H
      INTEGER :: NumEntries, NumHours, DayNum, Hours, Seq
      INTEGER :: OpenP, FilledP
      LOGICAL :: Good
      REAL    :: DVal, Tot, PrevP, RemP, EqP
      CHARACTER(LEN=4) :: S4
      CHARACTER(LEN=8) :: S8
      REAL, DIMENSION(:), ALLOCATABLE :: Precip
      
      !
      !  Set up the big array (1 hour bins)
      !  Array indices are 0 to NumHours-1
      !
      NumHours = (SeqFinish - SeqStart + 1) * 24
      ALLOCATE(Precip(0:NumHours-1), STAT=IOS)
      IF (IOS .NE. 0) THEN
         PRINT*, 'Error allocating RAM for sequential precipitation accumulation.'
         CALL EXIT(1)
      END IF
      Precip(:) = MissingData_Real
      
      !
      !  Apportion raw precip into the big precip array (1 hour bins)
      !
      NumEntries = UBOUND(RawData, 1)
      
      DO AH = 1, 48
         DO I = 1, NumEntries
            DO J = 1, 4
               Hours = RawData(I,J)%Hours
               IF (Hours .EQ. AH) THEN
                  DVal = RawData(I,J)%DVal
                  IF (DVal .GT. MissingData_Real_Test) THEN
                     WRITE(S8, 1000) RawData(I,J)%ObsDate
                     READ(S8, 1001) Y, M, D
                     WRITE(S4, 1004) RawData(I,J)%ObsTime
                     READ(S4, 1005) H
                     CALL DateSequence(D, M, Y, Seq)
                     H2 = (Seq-SeqStart)*24 + H        ! final hour for this precip accum
                     H1 = MAX(0, H2-AH+1)              ! first hour for this precip accum
                     PrevP = 0.0                       ! previously assigned precip in that period
                     FilledP = 0                       ! # hours already filled/assigned
                     DO H = H1, H2
                        IF (Precip(H) .GT. MissingData_Real_Test) THEN
                           PrevP = PrevP + Precip(H)
                           FilledP = FilledP + 1
                        END IF
                     END DO
                     OpenP = (H2-H1+1) - FilledP       ! # hours not filled (open for assignment)
                     IF (OpenP .GT. 0) THEN
                        RemP = DVal - PrevP               ! remaining precip to be apportioned
                        IF (RemP .LT. 0) RemP = 0.0
                        EqP = RemP / OpenP                ! equal precip amount to be assigned
                        DO H = H1, H2
                           IF (Precip(H) .LT. MissingData_Real_Test) THEN
                              Precip(H) = EqP
                           END IF
                        END DO
                     END IF
                  END IF
               END IF
            END DO
         END DO
      END DO
      
      !
      !  Now compute daily totals from the hourly bins
      !      
      DlyData(:) = MissingData_Real
      DO Seq = SeqStart, SeqFinish
         Tot = 0.0
         Good = .FALSE.
         DayNum = Seq - SeqStart + 1
         H1 = (DayNum-1)*24        ! first hour for this precip accum
         H2 = H1 + 23              ! final hour for this precip accum
         DO H = H1, H2
            IF (Precip(H) .GT. MissingData_Real_Test) THEN
               Tot = Tot + Precip(H)
               Good = .TRUE.
            END IF
         END DO
         IF (Good) DlyData(DayNum) = Tot
      END DO
      DEALLOCATE(Precip, STAT=IOS)
      
      
 1000 FORMAT(I8.8)
 1001 FORMAT(I4, 3I2)
 1004 FORMAT(I4.4, 4X)
 1005 FORMAT(I2)
  
      END SUBROUTINE ComputeDailyPrecip
  

!--------------------------------------------------------------------
      SUBROUTINE GetStationInfo(HistFileName, USAF_ID, WBAN_ID,    &
                 StnName, Country, StnLat, StnLong)
      IMPLICIT NONE
      CHARACTER(LEN=*), INTENT(IN)  :: HistFileName, USAF_ID, WBAN_ID
      CHARACTER(LEN=*), INTENT(OUT) :: StnName, Country
      REAL,             INTENT(OUT) :: StnLat, StnLong
                 
      INTEGER :: U1, IOS, I
      CHARACTER(LEN=5)  :: ID2
      CHARACTER(LEN=6)  :: ID1
      CHARACTER(LEN=59) :: RestOfLine

      StnName = '---'
      Country = '--'
      StnLat  = -999.999
      StnLong = -999.999
      
      U1 = GetFreeUnitNumber()
      OPEN(UNIT=U1, FILE=TRIM(HistFileName), STATUS='OLD', IOSTAT=IOS)
      IF (IOS .NE. 0) THEN
         PRINT*, 'Error opening the station history file. Process Aborted.'
         CALL EXIT(1)
      END IF
      CALL FileWasOpened(U1)

      READ(U1, 1000, IOSTAT=IOS) ID1, ID2, RestOfLine
      DO WHILE (IOS .EQ. 0)
         IF ((ID1 .EQ. USAF_ID) .AND. (ID2 .EQ. WBAN_ID)) THEN
            READ(RestOfLine, 1050, IOSTAT=I) StnName, Country, StnLat, StnLong
            IF (I .NE. 0) THEN
               StnName = '---'
               Country = '--'
               StnLat  = -999.999
               StnLong = -999.999
            END IF
            CLOSE(U1)
            CALL FileWasClosed(U1)
            RETURN            
         END IF
         READ(U1, 1000, IOSTAT=IOS) ID1, ID2, RestOfLine
      END DO
      CLOSE(U1)
      CALL FileWasClosed(U1)
      
     
 1000 FORMAT(A6, 1X, A5, 1X, A59)
 1050 FORMAT(A30, A2, 11X, F8.3, F8.3) 
 
      END SUBROUTINE GetStationInfo

!-----------------------------------------------------------------------------      
      SUBROUTINE ProcessOneStation(DataFileName, RetStat)
      IMPLICIT NONE
      CHARACTER(LEN=*), INTENT(IN)  :: DataFileName
      INTEGER,          INTENT(OUT) :: RetStat
      
      INTEGER :: I, J, U1, IOS, LN, NumLines
      INTEGER :: Year, Month, Day, HH, MM, NumDays
      INTEGER :: ErrStat, Seq, SeqStart, SeqFinish
      LOGICAL :: OK
      REAL    :: Latitude, Longitude, StnLat, StnLong
      CHARACTER(LEN=2)     :: Country
      CHARACTER(LEN=5)     :: WBAN_ID
      CHARACTER(LEN=6)     :: USAF_ID
      CHARACTER(LEN=30)    :: StnName
      CHARACTER(LEN=100)   :: FileName
      CHARACTER(LEN=10104) :: DataLine     ! max length is 105 mandatory plus up to 9999 variable 

      TYPE (DataFromOneLine)   :: DFOL
      TYPE (TMetCsvFile)       :: MetCsv
      !
      !  These arrays will be dimensioned (1:NumDays)
      !
      REAL, DIMENSION(:), ALLOCATABLE :: DlyTMax, DlyTMin, DlyPrec
      REAL, DIMENSION(:), ALLOCATABLE :: DlyAirT, DlyDewP, DlyWind, DlyClwd
      
      !
      !  These variables will be used for reading the raw data.
      !  Most are dimensioned (1:NumLines).
      !  Note that RawCloud is not actually raw (straight from file), but
      !  rather a result of analyzing the cloud layers on a single line.
      !
      !  The precip variable will be dimensioned (1:NumLines, 1:4) because
      !  there are, potentially, up to 4 precip observations on a single line
      !
      TYPE (ObsTimeValue), DIMENSION(:), ALLOCATABLE :: RawAirTemp, RawDewPt
      TYPE (ObsTimeValue), DIMENSION(:), ALLOCATABLE :: RawWind, RawCloud
      TYPE (PrecipObs), DIMENSION(:,:), ALLOCATABLE :: RawPrecip

      !
      !
      U1 = GetFreeUnitNumber()
      OPEN(UNIT=U1, FILE=TRIM(DataFileName), STATUS='OLD', ERR=811)
      CALL FileWasOpened(U1)
      
      !
      !  Read first line just to get the station ID
      !
      READ(U1, 1101, IOSTAT=IOS) USAF_ID, WBAN_ID, Latitude, Longitude
      IF (IOS .NE. 0) THEN
         PRINT*, 'Error reading station ID and location from line 1.'
         CALL EXIT(1)
      END IF
      
      !
      !  Do a fast pass through the file, determining the date extent
      !  and the total number of data lines. Data is ignored at this stage.
      !
      !  Note that I am translating the file time (UTC) into local time
      !  (Eastern Standard) for ALL stations, year-round. 
      !  Is that 100% correct?  Nope. But it's good enough for now.
      !  Later, this should probably be re-visited.
      !
      REWIND(U1)
      NumLines = 0
      CALL DateSequence(31, 12, 2999, SeqStart)
      CALL DateSequence( 1,  1, 1600, SeqFinish)
      READ(U1, 1102, IOSTAT=IOS) Year, Month, Day, HH, MM
      DO WHILE (IOS .EQ. 0) 
         NumLines = NumLines + 1
         CALL TranslateUTC_Local(Year, Month, Day, HH)
         CALL DateSequence(Day, Month, Year, Seq)
         IF (Seq .LT. SeqStart)  SeqStart  = Seq
         IF (Seq .GT. SeqFinish) SeqFinish = Seq
         READ(U1, 1102, IOSTAT=IOS) Year, Month, Day, HH, MM
      END DO
    
      !
      !  Set up the data storage
      !
      ALLOCATE(RawAirTemp(NumLines), RawDewpt(NumLines), RawWind(NumLines),     &
               RawCloud(NumLines), RawPrecip(NumLines,4), STAT=IOS)
      IF (IOS .NE. 0) THEN
         PRINT*, 'Error allocating the raw data storage arrays.'
         CALL EXIT(1)
      END IF
      
      NumDays = SeqFinish - SeqStart + 1
      ALLOCATE(DlyTMax(NumDays), DlyTMin(NumDays), DlyAirT(NumDays),       &
               DlyDewP(NumDays), DlyWind(NumDays), DlyClwd(NumDays),       &
               DlyPrec(NumDays), STAT=IOS)
      IF (IOS .NE. 0) THEN
         PRINT*, 'Error allocating the daily data storage arrays.'
         CALL EXIT(1)
      END IF
      
      !
      !  Initialize all of the data arrays
      !
      DO I = 1, NumLines
         RawAirTemp(I)%ObsDate = -99
         RawDewPt(I)%ObsDate = -99
         RawWind(I)%ObsDate = -99
         RawCloud(I)%ObsDate = -99
         DO J = 1, 4
            RawPrecip(I,J)%ObsDate = -99
         END DO
         
         RawAirTemp(I)%ObsTime = -99
         RawDewPt(I)%ObsTime = -99
         RawWind(I)%ObsTime = -99
         RawCloud(I)%ObsTime = -99
         DO J = 1, 4
            RawPrecip(I,J)%ObsTime = -99
         END DO

         RawAirTemp(I)%DVal = MissingData_Real
         RawDewPt(I)%DVal   = MissingData_Real
         RawWind(I)%DVal    = MissingData_Real
         RawCloud(I)%DVal   = MissingData_Real
         DO J = 1, 4
            RawPrecip(I,J)%DVal = MissingData_Real
         END DO
      END DO

      DlyTMax(:) = MissingData_Real
      DlyTMin(:) = MissingData_Real
      DlyAirT(:) = MissingData_Real
      DlyDewP(:) = MissingData_Real
      DlyWind(:) = MissingData_Real
      DlyClwd(:) = MissingData_Real
      DlyPrec(:) = MissingData_Real
      
      !
      !  Read and store raw data
      !
      REWIND(U1)
      LN = 1       ! Line Number
      READ(U1, 1105, IOSTAT=IOS) DataLine
      DO WHILE (IOS .EQ. 0)
         CALL ParseDataLine(DataLine, DFOL, ErrStat)
         OK = .TRUE.
         IF (ErrStat   .NE.       0) OK = .FALSE.
         IF (DFOL%USAF .NE. USAF_ID) OK = .FALSE.
         IF (DFOL%WBAN .NE. WBAN_ID) OK = .FALSE.
         IF (OK) THEN
            RawAirTemp(LN)%ObsDate = DFOL%ObsDate
            RawDewPt(LN)%ObsDate   = DFOL%ObsDate
            RawWind(LN)%ObsDate    = DFOL%ObsDate
            RawCloud(LN)%ObsDate   = DFOL%ObsDate

            RawAirTemp(LN)%ObsTime = DFOL%ObsTime
            RawDewPt(LN)%ObsTime   = DFOL%ObsTime
            RawWind(LN)%ObsTime    = DFOL%ObsTime
            RawCloud(LN)%ObsTime   = DFOL%ObsTime

            RawAirTemp(LN)%DVal = DFOL%AirTemp
            RawDewPt(LN)%DVal   = DFOL%DewPoint
            RawWind(LN)%DVal    = DFOL%WindSpd
            RawCloud(LN)%DVal   = DFOL%Cloud
            DO I = 1, 4
               RawPrecip(LN,I)%ObsDate = DFOL%ObsDate
               RawPrecip(LN,I)%ObsTime = DFOL%ObsTime
               RawPrecip(LN,I)%Hours   = DFOL%Precip(I)%Hours
               RawPrecip(LN,I)%DVal    = DFOL%Precip(I)%DVal
            END DO
         END IF
         LN = LN + 1
         READ(U1, 1105, IOSTAT=IOS) DataLine
      END DO
      
      CLOSE(U1)
      CALL FileWasClosed(U1)
      
      !
      !  Now that the raw data is read, turn it into daily values.
      !  Precip is handled very differently from the others.
      !
      CALL ComputeDailyMax(RawAirTemp, DlyTMax, SeqStart, SeqFinish)
      CALL ComputeDailyMin(RawAirTemp, DlyTMin, SeqStart, SeqFinish)
      CALL ComputeDailyAvg(RawAirTemp, DlyAirT, SeqStart, SeqFinish)
      CALL ComputeDailyAvg(RawDewPt,   DlyDewP, SeqStart, SeqFinish)
      CALL ComputeDailyAvg(RawWind,    DlyWind, SeqStart, SeqFinish)
      CALL ComputeDailyAvg(RawCloud,   DlyClwd, SeqStart, SeqFinish)

      CALL ComputeDailyPrecip(RawPrecip, DlyPrec, SeqStart, SeqFinish)

      !
      !  Build/populate the data structure used for writing the output file
      !
      CALL InitializeDlyMetStnDataset(MetCsv%DlyData)

      CALL GetStationInfo('ish-history.txt', USAF_ID, WBAN_ID, StnName,  &
                           Country, StnLat, StnLong)
      
      MetCsv%DlyData%ID      = TRIM(USAF_ID) // '-' // TRIM(WBAN_ID)
      MetCsv%DlyData%Name    = TRIM(StnName)
      MetCsv%DlyData%Country = TRIM(Country)
      MetCsv%DlyData%Lat     = StnLat
      MetCsv%DlyData%Long    = StnLong
      MetCsv%DlyData%SDate   = SeqStart
      MetCsv%DlyData%EDate   = SeqFinish
      MetCsv%DlyData%NumDataTypes = 7

      ALLOCATE(MetCsv%DlyData%DataTypes(7), MetCsv%DlyData%DataUnits(7),      &
               MetCsv%DlyData%MetData(7,NumDays), STAT=IOS)
      IF (IOS .NE. 0) THEN
         PRINT*, 'Error allocating RAM for the output data.'
         CALL EXIT(1)
      END IF
      
      MetCsv%DlyData%DataTypes(1) = MDT_AirtempMax
      MetCsv%DlyData%DataTypes(2) = MDT_AirtempMin
      MetCsv%DlyData%DataTypes(3) = MDT_Precipitation
      MetCsv%DlyData%DataTypes(4) = MDT_AirtempMean
      MetCsv%DlyData%DataTypes(5) = MDT_DewpointMean
      MetCsv%DlyData%DataTypes(6) = MDT_WindSpeed
      MetCsv%DlyData%DataTypes(7) = MDT_CloudCover

      MetCsv%DlyData%DataUnits(1) = MDU_Celsius
      MetCsv%DlyData%DataUnits(2) = MDU_Celsius
      MetCsv%DlyData%DataUnits(3) = MDU_Millimeters
      MetCsv%DlyData%DataUnits(4) = MDU_Celsius
      MetCsv%DlyData%DataUnits(5) = MDU_Celsius
      MetCsv%DlyData%DataUnits(6) = MDU_MetersPerSecond
      MetCsv%DlyData%DataUnits(7) = MDU_Percent
      
      DO Seq = SeqStart, SeqFinish
         I = Seq - SeqStart + 1
         MetCsv%DlyData%MetData(1,I) = DlyTMax(I)
         MetCsv%DlyData%MetData(2,I) = DlyTMin(I)
         MetCsv%DlyData%MetData(3,I) = DlyPrec(I)
         MetCsv%DlyData%MetData(4,I) = DlyAirT(I)
         MetCsv%DlyData%MetData(5,I) = DlyDewP(I)
         MetCsv%DlyData%MetData(6,I) = DlyWind(I)
         MetCsv%DlyData%MetData(7,I) = DlyClwd(I)
      END DO      

      !
      !  Write the CSV-format output file
      !
      FileName = 'met_' // TRIM(USAF_ID) // '-' // TRIM(WBAN_ID) // '.csv'
      CALL WriteMetCsvFile(TRIM(FileName), MetCsv, IOS)
      IF (IOS .NE. 0) THEN
         PRINT*, 'Error writing output CSV file. Aborting.'
         CALL EXIT(1)
      END IF

      ErrorLevel = 0   
      GOTO 999     ! clean up and exit
      
      !
      !  Error handling
      !
  811 ErrorMessage = 'Error opening input file '//TRIM(DataFileName)
      CALL PassMsg; GOTO 898
      
  898 ErrorLevel = 1
  899 ErrorMessage = 'Program aborted!';  CALL PassMsg


  999 DEALLOCATE(MetCsv%DlyData%MetData, MetCsv%DlyData%DataUnits,       &
                 MetCsv%DlyData%DataTypes, STAT=IOS)
      DEALLOCATE(DlyTMax, DlyTMin, DlyPrec, DlyAirT, DlyDewP,      &
                 DlyWind, DlyClwd, STAT=IOS)
      DEALLOCATE(RawAirTemp, RawDewpt, RawWind,                    &
                 RawCloud, RawPrecip, STAT=IOS)
      RetStat = ErrorLevel
      
      
 1101 FORMAT(4X, A6, A5, 13X, F6.2, F7.2)
 1102 FORMAT(15X, I4, 4I2)
 1105 FORMAT(A10104)
      
      END SUBROUTINE ProcessOneStation      
      
      
   END MODULE Process_ISH_Subs
   
   
!==============================================================================   
!==============================================================================   
!==============================================================================   
   PROGRAM Process_ISH
   USE Process_ISH_Subs
   IMPLICIT NONE

   INTEGER :: I, ArgC
   INTEGER :: ErrStat
   CHARACTER(LEN=100)  :: DataFileName, Args(1000)
   CHARACTER(LEN=9999) :: CLine
   
   !
   !  Get the data file name from the command line
   !
   CALL GET_COMMAND(CLine)
   CALL ParseCmdLine(CLine, Args, ArgC)
   IF (ErrorLevel .NE. 0) ArgC = 0
   IF (ArgC .LT. 2) THEN
      PRINT*, 'USAGE: Process_ISH  filename(s)'
      PRINT*, '     filename(s) = data file(s) containing ish-format dataset (wildcards ok)'
      PRINT*, 'Note that ish-history.txt must also exist in the current folder.'
      CALL EXIT(1)
   END IF
   
   DO I = 2, ArgC
      DataFileName = Args(I)
      print*, 'Processing '//TRIM(DataFileName)
      CALL ProcessOneStation(DataFileName, ErrStat)
      IF (ErrStat .NE. 0) THEN
         PRINT*, 'Program aborted.'
         CALL EXIT(1)
      END IF
   END DO
   
   END PROGRAM