!-----------------------------------------------------------------------
!  This module provides routines for reading and writing to/from a
!  met_*.csv file, the new (2014) format that I am using as the GLERL
!  standard format for met station data.  
!
!  For more details, see the written file format documentation that should
!  be available elsewhere.
!
!  Notes:
!
!  1) The file format as I defined it allows for the data to be stored 
!  in a variety of units, but the preferred units are what will be 
!  used for the write procedures in this module. The read procedures
!  will handle whatever is thrown at them (within the limits of those I
!  have anticipated).
!
!  2) When writing the files, I will always output the same 7 data types
!  in the same prescribed order. This is not required by the format 
!  definition , but will make writing easier.  The read procedures are
!  accommodating to more variance in the data. Columns may be missing and
!  they may be in a different order.
!
!-----------------------------------------------------------------------

MODULE ReadWriteMetCsv
   USE ErrorProcess
   USE GlerlUtil
   USE MetDataTypesAndUnits
   USE DailyMetStationDataset


   TYPE TMetCsvFile
      CHARACTER(LEN=150)       :: FileName
      TYPE (TDlyMetStnDataset) :: DlyData
   END TYPE TMetCsvFile
   
CONTAINS

   !------------------------------------------------------------
   SUBROUTINE ReadMetCsvFile(FileName, MCF, ErrStat)
   IMPLICIT NONE
   CHARACTER(LEN=*),    INTENT(IN)    :: FileName
   TYPE (TMetCsvFile),  INTENT(INOUT) :: MCF
   INTEGER,             INTENT(OUT)   :: ErrStat

   INTEGER :: I, J, K, IOS, UNum
   INTEGER :: IOSYr, IOSMn, IOSDy, IOSX
   INTEGER :: Year, Month, Day, NumDays, Seq
   INTEGER :: NumDT, NumDU, LineNum
   REAL    :: DVal
   CHARACTER (LEN=100) :: TStr
   CHARACTER (LEN=250) :: Line

   INTEGER, DIMENSION(1) :: UB
   CHARACTER(LEN=100), DIMENSION(99) :: CsvStrings
   
   !
   !  Initialize the data structure   
   !
   CALL InitializeDlyMetStnDataset(MCF%DlyData)
   MCF%FileName = TRIM(FileName)
   print*, 'Input Filename=['//TRIM(FileName)//']'
   print*, '1)MCF%Filename=['//TRIM(MCF%FileName)//']'
   
   !
   !  Read header info so I can allocate RAM, etc
   !
   ErrStat = -1
   CALL ReadMetCsvFileHeaderInfo(FileName, MCF, ErrStat)
   IF (ErrStat .NE. 0) GOTO 899
   NumDT = MCF%DlyData%NumDataTypes
   print*, '2)MCF%Filename=['//TRIM(MCF%FileName)//']'

   !
   !  Allocate the RAM.
   !  Note that the DataTypes and DataUnits arrays were already allocated 
   !  and filled as part of reading the header info.
   !  Extract all of the data.
   !
   NumDays = MCF%DlyData%EDate - MCF%DlyData%SDate + 1
   ALLOCATE(MCF%DlyData%MetData(NumDT, NumDays), STAT=IOS)
   IF (IOS .NE. 0) THEN
      ErrorMessage = 'Error when allocating RAM for data values.'
      CALL PassMsg
      WRITE(ErrorMessage, *) 'NumDT = ', NumDT
      CALL PassMsg
      WRITE(ErrorMessage, *) 'NumDays = ', NumDays
      CALL PassMsg
      CALL InitializeDlyMetStnDataset(MCF%DlyData)
      ErrorLevel = 1
      GOTO 899
   END IF
   MCF%DlyData%MetData(:,:) = MissingData_Real

   !
   !  Read all of the data.
   !   
   UNum = GetFreeUnitNumber();  IF (ErrorLevel .NE. 0) GOTO 899
   OPEN(UNIT=UNum, FILE=TRIM(FileName), STATUS='OLD', ERR=811)
   CALL FileWasOpened(UNum)
   DO I = 1, 6
      READ(UNum, 1000, IOSTAT=IOS)      ! skip header lines
   END DO
   
   LineNum = 7
   READ(UNum, 1000, IOSTAT=IOS) Line
   DO WHILE (IOS .EQ. 0) 
      CALL ParseCommaSepLine(Line, CsvStrings)
      TStr = TRIM(CsvStrings(1)); READ(TStr, *, IOSTAT=IOSYr) Year
      TStr = TRIM(CsvStrings(2)); READ(TStr, *, IOSTAT=IOSMn) Month
      TStr = TRIM(CsvStrings(3)); READ(TStr, *, IOSTAT=IOSDy) Day
      IF ((IOSYr+IOSMn+IOSDy) .NE. 0) THEN
         WRITE(ErrorMessage, 1001) LineNum, TRIM(FileName)
         CALL PassMsg
         CALL InitializeDlyMetStnDataset(MCF%DlyData)
         ErrorLevel = 1
         GOTO 899
      ELSE
         CALL DateSequence(Day, Month, Year, Seq)
         IF ((Seq .LT. MCF%DlyData%SDate) .OR. (Seq .GT. MCF%DlyData%EDate)) THEN
            !  maybe give some feedback here????
         ELSE
            J = Seq - MCF%DlyData%SDate + 1
            DO I = 1, NumDT
               TStr = TRIM(CsvStrings(3+I)); READ(TStr, *, IOSTAT=IOSX) DVal
               IF (IOSX .EQ. 0) THEN
                  MCF%DlyData%MetData(I,J) = DVal
               ELSE
                  MCF%DlyData%MetData(I,J) = MissingData_Real
                  ! feedback???
               END IF
            END DO
         END IF
      END IF
      LineNum = LineNum + 1
      READ(UNum, 1000, IOSTAT=IOS) Line
   END DO   
   CLOSE(UNum)
   CALL FileWasClosed(UNum)
   print*, '3)MCF%Filename=['//TRIM(MCF%FileName)//']'

   !
   !  Future enhancement here....
   !  Trim the data structure so that the data extents correspond to only
   !  the NON-MISSING data. That includes adjusting SDate and EDate.
   !
   ErrStat = 0
   RETURN
   

 811 ErrorMessage = 'Error opening input file '//TRIM(FileName); CALL PassMsg; GOTO 899
 
 899 ErrorMessage = '[traceback] : ReadMetCsvFile'; CALL PassMsg
     RETURN

 1000 FORMAT(A)
 1001 FORMAT('Error parsing date on line ', I0, ' in file ', A)
     

   END SUBROUTINE ReadMetCsvFile

   !------------------------------------------------------------
   SUBROUTINE ReadMetCsvFileHeaderInfo(FileName, MCF, ErrStat)
   IMPLICIT NONE
   
   CHARACTER(LEN=*),   INTENT(IN)    :: FileName
   TYPE (TMetCsvFile), INTENT(INOUT) :: MCF
   INTEGER,            INTENT(OUT)   :: ErrStat

   INTEGER :: I, IOS, UNum
   INTEGER :: Year, Month, Day, NumDays
   INTEGER :: NumStrings, NumDT, NumDU
   CHARACTER(LEN=100) :: TStr
   CHARACTER(LEN=250) :: Line
   
   INTEGER, DIMENSION(1) :: UB
   CHARACTER(LEN=100), DIMENSION(99) :: CsvStrings

   !
   !  Initialize stuff
   !
   CALL ClearDlyMetStnDatasetHeader(MCF%DlyData)
   ErrStat = -1

   !
   !  Open the input file
   !
   UNum = GetFreeUnitNumber();  IF (ErrorLevel .NE. 0) GOTO 899
   OPEN(UNIT=UNum, FILE=TRIM(FileName), STATUS='OLD', ERR=811)
   CALL FileWasOpened(UNum)
   
   !
   !  1) Station ID and Name
   !
   READ(UNum, 1000, ERR=812) Line
   CALL ParseCommaSepLine(Line, CsvStrings, .TRUE., NumStrings)
   IF (NumStrings .LT. 2) GOTO 601
   MCF%DlyData%ID   = TRIM(ADJUSTL(CsvStrings(1)))
   MCF%DlyData%Name = TRIM(ADJUSTL(CsvStrings(2)))
   IF (NumStrings .GT. 2) MCF%DlyData%Country = TRIM(ADJUSTL(CsvStrings(3)))
   
   !
   !  2) Station Latitude and Longitude
   !
   READ(UNum, 1000, ERR=812) Line
   CALL CAPS(Line)
   CALL ParseCommaSepLine(Line, CsvStrings, .TRUE., NumStrings)
   IF (NumStrings .LT. 3) GOTO 602
   IF (TRIM(CsvStrings(1)) .NE. 'LAT&LONG') GOTO 752
   TStr = TRIM(ADJUSTL(CsvStrings(2)));  READ(TStr, *, ERR=701) MCF%DlyData%Lat
   TStr = TRIM(ADJUSTL(CsvStrings(3)));  READ(TStr, *, ERR=702) MCF%DlyData%Long
   
   !
   !  3) Data start date
   !
   READ(UNum, 1000, ERR=812) Line
   CALL CAPS(Line)
   CALL ParseCommaSepLine(Line, CsvStrings, .TRUE., NumStrings)
   IF (NumStrings .LT. 3) GOTO 603
   IF  (TRIM(CsvStrings(1)) .NE. 'STARTS(YMD)') GOTO 753
   TStr = TRIM(ADJUSTL(CsvStrings(2))); READ(TStr, *, ERR=703) Year
   TStr = TRIM(ADJUSTL(CsvStrings(3))); READ(TStr, *, ERR=703) Month
   TStr = TRIM(ADJUSTL(CsvStrings(4))); READ(TStr, *, ERR=703) Day
   CALL DateSequence(Day, Month, Year, MCF%DlyData%SDate)
   IF (ErrorLevel .NE. 0) GOTO 703
   
   !
   !  4) Data end date
   !
   READ(UNum, 1000, ERR=812) Line
   CALL CAPS(Line)
   CALL ParseCommaSepLine(Line, CsvStrings, .TRUE., NumStrings)
   IF (NumStrings .LT. 3) GOTO 604
   IF  (TRIM(CsvStrings(1)) .NE. 'ENDS(YMD)') GOTO 754
   TStr = TRIM(ADJUSTL(CsvStrings(2))); READ(TStr, *, ERR=703) Year
   TStr = TRIM(ADJUSTL(CsvStrings(3))); READ(TStr, *, ERR=703) Month
   TStr = TRIM(ADJUSTL(CsvStrings(4))); READ(TStr, *, ERR=703) Day
   CALL DateSequence(Day, Month, Year, MCF%DlyData%EDate)
   IF (ErrorLevel .NE. 0) GOTO 704
   
   NumDays = MCF%DlyData%EDate - MCF%DlyData%SDate + 1
   IF (NumDays .LT. 1) THEN
      ErrorMessage = 'Invalid start/end dates in file: '//TRIM(FileName)
      CALL PassMsg
      ErrorMessage = 'End date prior to start date'; CALL PassMsg
      GOTO 899
   END IF
   
   !
   !  5) Data types
   !     NumDT = number of data types (i.e data columns) in the file
   !
   READ(UNum, 1000, ERR=812) Line
   CALL CAPS(Line)
   CALL ParseCommaSepLine(Line, CsvStrings, .TRUE., NumStrings)
   IF (NumStrings .LT. 3) GOTO 605
   IF (TRIM(CsvStrings(1)) .NE. 'DATE') GOTO 755
   IF (TRIM(CsvStrings(2)) .NE. 'DATE') GOTO 755
   IF (TRIM(CsvStrings(3)) .NE. 'DATE') GOTO 755
   NumDT = NumStrings - 3
   MCF%DlyData%NumDataTypes = NumDT
   ALLOCATE(MCF%DlyData%DataTypes(NumDT), STAT=IOS)
   MCF%DlyData%DataTypes(:) = MDT_Undefined
   DO I = 4, NumStrings
      TStr = TRIM(ADJUSTL(CsvStrings(I)))
      MCF%DlyData%DataTypes(I-3) = MetDataTypeFromString(TRIM(TStr))
   END DO

   !
   !  6) Data units for each column in file
   !
   READ(UNum, 1000, ERR=812) Line
   CALL CAPS(Line)
   CALL ParseCommaSepLine(Line, CsvStrings, .TRUE., NumStrings)
   IF (NumStrings .LT. 3) GOTO 606
   IF (TRIM(CsvStrings(1)) .NE. 'YYYY') GOTO 756
   IF (TRIM(CsvStrings(2)) .NE. 'MM')   GOTO 756
   IF (TRIM(CsvStrings(3)) .NE. 'DD')   GOTO 756
   NumDU = NumStrings - 3
   IF (NumDU .NE. NumDT) GOTO 610
   ALLOCATE(MCF%DlyData%DataUnits(NumDU), STAT=IOS)
   MCF%DlyData%DataUnits(:) = MDU_Undefined
   DO I = 4, NumStrings
      TStr = TRIM(ADJUSTL(CsvStrings(I)))
      MCF%DlyData%DataUnits(I-3) = MetDataUnitFromString(TRIM(TStr))
   END DO

   CLOSE(UNum)
   CALL FileWasClosed(UNum)
   
   ErrStat = 0
   RETURN
   
   
 601 ErrorMessage = 'Formatting error on line 1 (too few entries) in '//TRIM(FileName); CALL PassMsg; GOTO 899
 602 ErrorMessage = 'Formatting error on line 2 (too few entries) in '//TRIM(FileName); CALL PassMsg; GOTO 899
 603 ErrorMessage = 'Formatting error on line 3 (too few entries) in '//TRIM(FileName); CALL PassMsg; GOTO 899
 604 ErrorMessage = 'Formatting error on line 4 (too few entries) in '//TRIM(FileName); CALL PassMsg; GOTO 899
 605 ErrorMessage = 'Formatting error on line 5 (too few entries) in '//TRIM(FileName); CALL PassMsg; GOTO 899
 606 ErrorMessage = 'Formatting error on line 6 (too few entries) in '//TRIM(FileName); CALL PassMsg; GOTO 899
 610 ErrorMessage = 'Error: Mismatched entry count between data types and data units '//TRIM(FileName); CALL PassMsg; GOTO 899
   
 701 ErrorMessage = 'Error parsing Latitude on line 2 in '//TRIM(FileName); CALL PassMsg; GOTO 899
 702 ErrorMessage = 'Error parsing Longitude on line 2 in '//TRIM(FileName); CALL PassMsg; GOTO 899
 703 ErrorMessage = 'Error parsing start date on line 3 in '//TRIM(FileName); CALL PassMsg; GOTO 899
 704 ErrorMessage = 'Error parsing end date on line 4 in '//TRIM(FileName); CALL PassMsg; GOTO 899
   
 752 ErrorMessage = 'Invalid format for line 2 in '//TRIM(FileName); CALL PassMsg; GOTO 899
 753 ErrorMessage = 'Invalid format for line 3 in '//TRIM(FileName); CALL PassMsg; GOTO 899
 754 ErrorMessage = 'Invalid format for line 4 in '//TRIM(FileName); CALL PassMsg; GOTO 899
 755 ErrorMessage = 'Invalid format for line 5 in '//TRIM(FileName); CALL PassMsg; GOTO 899
 756 ErrorMessage = 'Invalid format for line 5 in '//TRIM(FileName); CALL PassMsg; GOTO 899
  
 811 ErrorMessage = 'Error opening input file '//TRIM(FileName); CALL PassMsg; GOTO 899
 812 ErrorMessage = 'Error reading input file '//TRIM(FileName); CALL PassMsg; GOTO 899
   
   
 899 ErrorMessage = '[traceback] : ReadFileHeaderInfo'; CALL PassMsg
     DEALLOCATE(MCF%DlyData%DataTypes, STAT=IOS)
     DEALLOCATE(MCF%DlyData%DataUnits, STAT=IOS)
     RETURN

 1000 FORMAT(A)
     
   END SUBROUTINE ReadMetCsvFileHeaderInfo

   !------------------------------------------------------------
   SUBROUTINE WriteMetCsvFile(FileName, MCF, ErrStat)
   IMPLICIT NONE
   CHARACTER(LEN=*),    INTENT(IN)    :: FileName
   TYPE (TMetCsvFile),  INTENT(INOUT) :: MCF
   INTEGER,             INTENT(OUT)   :: ErrStat

   INTEGER :: I, J, IOS, UNum, Dy, Mn, Yr, Seq, NDT
   REAL    :: DVal
   CHARACTER(LEN=1)   :: Comma, DblQ
   CHARACTER(LEN=30)  :: S
   CHARACTER(LEN=999) :: Line
   CHARACTER(LEN=100), DIMENSION(50) :: CsvStrings
   
   Comma = ','
   DblQ  = '"'
   
   !
   !  Open the output file
   !
   UNum = GetFreeUnitNumber();  IF (ErrorLevel .NE. 0) GOTO 899
   OPEN(UNIT=UNum, FILE=TRIM(FileName), STATUS='REPLACE', ERR=811)
   CALL FileWasOpened(UNum)

   !
   !  
   !
   WRITE(UNum, 1001) DblQ, TRIM(MCF%DlyData%ID), DblQ, Comma, DblQ, TRIM(MCF%DlyData%Name), DblQ
   WRITE(UNum, 1002) DblQ, 'Lat&Long', DblQ, Comma, MCF%DlyData%Lat, Comma, MCF%DlyData%Long

   IF (MCF%DlyData%SDate .NE. DateSeq_MissingDate) THEN
      CALL SequenceDate(Dy, Mn, Yr, MCF%DlyData%SDate)
   ELSE
      Dy=-99; Mn=-99; Yr = -9999
   ENDIF
   WRITE(UNum, 1003) DblQ, 'Starts(YMD)', DblQ, Comma, Yr, Comma, Mn, Comma, Dy
   
   IF (MCF%DlyData%EDate .NE. DateSeq_MissingDate) THEN
      CALL SequenceDate(Dy, Mn, Yr, MCF%DlyData%EDate)
   ELSE
      Dy=-99; Mn=-99; Yr = -9999
   ENDIF
   WRITE(UNum, 1003) DblQ, 'Ends(YMD)', DblQ, Comma, Yr, Comma, Mn, Comma, Dy
   
   NDT = MCF%DlyData%NumDataTypes
   Line = ''
   DO I = 1, 3
      Line = TRIM(Line) // DblQ//'Date'//DblQ//Comma
   END DO
   DO I = 1, NDT-1
      S = DblQ // TRIM(MetDataTypeString(MCF%DlyData%DataTypes(I))) // DblQ
      Line = TRIM(Line) // TRIM(S) // Comma
   END DO
   S = DblQ // TRIM(MetDataTypeString(MCF%DlyData%DataTypes(NDT))) // DblQ
   Line = TRIM(Line) // TRIM(S)
   WRITE(UNum, 1005) TRIM(Line)

   NDT = MCF%DlyData%NumDataTypes
   Line = DblQ//'YYYY'//DblQ//Comma
   Line = TRIM(Line) // DblQ//'MM'//DblQ//Comma
   Line = TRIM(Line) // DblQ//'DD'//DblQ//Comma
   DO I = 1, NDT-1
      S = DblQ // TRIM(MetDataUnitString(MCF%DlyData%DataUnits(I))) // DblQ
      Line = TRIM(Line) // TRIM(S) // Comma
   END DO
   S = DblQ // TRIM(MetDataUnitString(MCF%DlyData%DataUnits(NDT))) // DblQ
   Line = TRIM(Line) // TRIM(S)
   WRITE(UNum, 1005) TRIM(Line)
                  
   !
   !  Data value rows
   !  Note that for each valid numeric value, I want to make sure there is at
   !  least 1 digit in front of the decimal point. Default Fortran formatting,
   !  at least with the gfortran compiler, is annoying in that a number like
   !  0.75 will end up printing as .75 and -0.75 will print as -.75
   !  While this is ok in terms of validity, it is not "pretty" from a user
   !  standpoint. It also seems like it caused issues for me in the past when
   !  trying to read the resulting file with code from other languages (not
   !  100% sure about that one). Anyway, forcing the leading 0 is the best
   !  universal solution in my mind. So I do some messy manipulation of the
   !  output data values in order to accomplish that.
   !
   !  Also note that I am employing the MS Excel convention of #N/A to indicate
   !  missing data values. It allows very clean input to Excel, disambiguation of
   !  what value means missing, and a user of the resulting file can do a global 
   !  replace if they don't like it.
   !
   !  Additionally, due to the way Fortran formatting is implemented, I have to do
   !  special stuff to avoid a trailing comma on the last entry. You would think it
   !  could be done by specifying it as a leading comma and structuring things
   !  appropriately, it doesn't work quite as you might expect. So I am stuck with
   !  a lot of extra messiness, unfortunately. This is a bad omission in the Fortran
   !  standard (missing any way to force leading zeroes in a REAL output) that I saw
   !  discussed online numerous times from people trying to do similar stuff.
   !
   NDT = MCF%DlyData%NumDataTypes
   DO Seq = MCF%DlyData%SDate, MCF%DlyData%EDate
      I = Seq - MCF%DlyData%SDate + 1
      CALL SequenceDate(Dy, Mn, Yr, Seq)
      Line = ''
      DO J = 1, NDT
         DVal = MCF%DlyData%MetData(J,I)
         IF (IsMissing(DVal)) THEN
            S = '#N/A'
         ELSE
            IF (ABS(DVal) .GT. 1.0) THEN
               WRITE(S, 1050) DVal
            ELSE
               IF (DVal .GE. 0) THEN
                  WRITE(S, 1051) INT(DVal), DVal-INT(DVal)
               ELSE
                  DVal = ABS(DVal)
                  WRITE(S, 1052) INT(DVal), DVal-INT(DVal)
               END IF
            END IF
         END IF
         IF (J .LT. NDT) S = TRIM(S) // ','
         Line = TRIM(Line) // TRIM(S)
      END DO
      WRITE(UNum, 1010) Yr, Mn, Dy, TRIM(Line)
   END DO
   
   CLOSE(UNum)
   CALL FileWasClosed(UNum)
   ErrStat = 0
   RETURN
   
 811 ErrorMessage = 'Error opening output file '//TRIM(FileName); CALL PassMsg; GOTO 899

 899 ErrorMessage = '[traceback] : WriteMetCsvFile'; CALL PassMsg
     ErrStat = 1
     RETURN
   
   
 1001 FORMAT(A1,A,A1, A1, A1, A, A1)
 1002 FORMAT(A1,A,A1,A1, F0.3, A1, F0.3)
 1003 FORMAT(A1,A,A1,A1, I0, A1, I0, A1, I0)
 1005 FORMAT(A)
 1010 FORMAT(I4.4, 2(',',I2.2), ',', A)
 1050 FORMAT(F0.3)
 1051 FORMAT(I0.1, F0.3)
 1052 FORMAT('-', I0.1, F0.3)
 
 
   END SUBROUTINE WriteMetCsvFile


END MODULE ReadWriteMetCsv